/*
  Model-View-Controller demo program.
  Copyright (C) 2006 by Maria Litvin and Gary Litvin.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details:
    http://www.gnu.org/licenses/gpl.html.
*/

// Represents a model for a vehicle moving along a straight line.
// Serves as a "model" in the MVC demo.
public class Car extends java.util.Observable
{
  private static final double G = 9.8;  // acceleration due to gravity m/sec^2
  private final double maxSpeed = 120 / 2.2;   // meters/sec = 120 mi/hr
  private final double minSpeed =  12 / 2.2;   // meters/sec =  12 mi/hr

  // Motion model coefficients (see the move method):
  private final double slowDown = 1.0;
  private final double gasFactor = 0.1; 
  private final double brakeFactor = 0.1;
  private final double windFactor = 0.625;
 
  // Car data fields:
  private double position = 0.0;      // meters
  private double speed = 0.0;         // meters/sec
  private double acceleration = 0.0;  // meters/sec^2
  private boolean brakesLocked = false;
  private double gasValue;
  private double brakeValue;

  // Constructs a new car, which is not moving.
  public Car()
  {
    gasValue = 0.0;
    brakeValue = 0.0;
    position = 0;
  }  

  // Returns the position of this car (in meters).
  public double getPosition()
  {
    return position;
  }

  // Returns the speed of this car (in meters/sec).
  public double getSpeed()
  {
    return speed;
  }

  // Returns the acceleration of this car (in meters/sec^2).
  public double getAcceleration()
  {
    return acceleration;
  }

  // Sets the value from the gas pedal.
  public void applyGas(int value)
  {
    gasValue = value;
  }

  // Sets the value from the brake pedal.
  public void applyBrakes(int value)
  {
    brakeValue = value;
  }

  // Stops this car.
  private void stop()
  {
    speed = 0;
    acceleration = 0;
  }

  // Returns true if this car was braking too fast and its
  // brakes got locked; otherwise returns false.
  public boolean roughStop()
  {
    return brakesLocked;
  }

  // Increments or decrements the speed of this car and increments its position
  // appropriately for the time interval dt.
  public void move(double dt)
  {
    acceleration =  gasFactor * gasValue *
                        (1.0 - windFactor * speed / maxSpeed) -
                    brakeFactor * brakeValue - slowDown;
    if (acceleration >= 0)
      brakesLocked = false;
    else if (acceleration < -G)
      brakesLocked = true;

    speed += acceleration * dt;
    if (speed < minSpeed && acceleration < 0)
      stop();
    if (speed > maxSpeed)
      speed = maxSpeed;

    position += speed * dt;
    if (position > Double.MAX_VALUE / 2)  // to avoid double overflow
      position = 0;

    setChanged();
    notifyObservers();
  }
}
